//
// Created by hangshu on 22-4-26.
//

#ifndef SYSY2022_BJTU_BASICBLOCK_HH
#define SYSY2022_BJTU_BASICBLOCK_HH
#include <string>
#include "Instruction.hh"
#include <list>
#include <set>
#include "instr.hh"
class BasicBlock{
public:
    std::vector<Instruction*> ir;
    BasicBlock* parent;
    std::vector<Value*> vars;
    std::string name;
    std::set<Value*> originVals;

    BasicBlock(std::string name):name(name){}
    BasicBlock(BasicBlock* parent,std::string func_name,int cnt) {
        this->parent = parent;
        name = func_name + "::BB" +  std::to_string(cnt);
    }
    void pushIr(Instruction* instruction);
    std::vector<Instruction*>& getIr();
    void pushVar(Value *v);
    void insertDomFrontier(BasicBlock* bb);
    std::set<BasicBlock*> getDomFrontier();
    virtual void print(std::ostream& out) = 0;
    virtual void clear() = 0;
    virtual ~BasicBlock(){}
    void pushPre(BasicBlock* nb);
    void pushSucc(BasicBlock* nb);
    void setPre(std::vector<BasicBlock*> pre);
    void setSucc(std::vector<BasicBlock*> succ);
    std::vector<BasicBlock*> getPre();
    std::vector<BasicBlock*> getSucc();
    void setIdom(BasicBlock* idom);
    BasicBlock* getIdom() ;
    void pushDomTreeSuccNode(BasicBlock* bb) ;
    std::vector<BasicBlock*> getDomTreeSuccNode();
    //add for codegen
    void pushInstr(Instr* instr);
    std::vector<Instr*>& getInstrs();
private:
    std::vector<BasicBlock*> preBBs;
    std::vector<BasicBlock*> succBBs;
    std::vector<Instr*> instrs;
    std::set<BasicBlock*> domFrontier;
    BasicBlock* idom = nullptr;
    std::vector<BasicBlock*> domTreeSuccNode;
};
class NormalBlock:public BasicBlock{
public:
    BasicBlock* nextBB = nullptr;
    NormalBlock(BasicBlock* parent,std::string func_name, int cnt) : BasicBlock(parent,func_name,cnt){}
    NormalBlock(std::string name): BasicBlock(name){}
    void print(std::ostream& out) override final;
    void clear() override final{}

};
class CondBlock:public BasicBlock{
public:
    BasicBlock* trueBB = nullptr;
    BasicBlock* falseBB = nullptr;
    TempVal val;
    bool isAnd = false;
    CondBlock(BasicBlock* parent,std::string func_name,int cnt): BasicBlock(parent,func_name,cnt){}
    void print(std::ostream& out);
    void clear() override final{}
};
class SelectBlock:public BasicBlock{
public:
    std::vector<BasicBlock*> cond;
    std::vector<BasicBlock*> ifStmt;
    std::vector<BasicBlock*> elseStmt;

    SelectBlock(BasicBlock* parent,std::string func_name,int cnt) : BasicBlock(parent,func_name,cnt){}
    void print(std::ostream& out) override final;
    void clear() override final;
};
class IterationBlock:public BasicBlock{
public:
    std::vector<BasicBlock*> cond;
    std::vector<BasicBlock*> whileStmt;

    IterationBlock(BasicBlock* parent,std::string func_name,int cnt) : BasicBlock(parent,func_name,cnt){}
    void print(std::ostream& out) override final;
    void clear() override final;
};
#endif //SYSY2022_BJTU_BASICBLOCK_HH